/*
 * HTGeodesic.java
 * www.bouthier.net
 *
 * The MIT License :
 * -----------------
 * Copyright (c) 2001 Christophe Bouthier
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

// Cdigo alterado por Silvio Evangelista

//package hypertree;

import java.awt.Color;
import java.awt.Graphics;

/**
 * The HTGeodesic class implements a geodesic 
 * linking to points in the Poincarre model.
 */
class HTGeodesic {

    private static final double EPSILON = 1.0E-5; // epsilon

    private static final int    LINE    = 0;       // draw a line
    private static final int    ARC     = 1;       // draw an arc

    private int      type = LINE; // type of the geodesic

    private HTCoordE npoint = null;
    private HTCoordE za   = null; // first point (Euclidian)
    private HTCoordE zb   = null; // second point (Euclidian)
    private HTCoordE zc   = null; // control point (Euclidian)
    private HTCoordE zo   = null; // center of the geodesic;

    private HTCoordS a    = null; // first point (on the screen)
    private HTCoordS b    = null; // second point (on the screen)
    private HTCoordS c    = null; // control point (on the screen)

    private int numberSubintervals = 11;
    private int numberHandles = 3; // number points to interpolate... e.g., 2

    private double[] xcoord = new double[numberHandles];
    private double[] ycoord = new double[numberHandles]; // points to interpolate  
    private double[] yValues = new double[numberSubintervals + 1];
    private double[] xValues = new double[numberSubintervals + 1];
    private Color color;
    private Color colorNode= Color.blue;
    private double PA, PB, PC;
      
    /* --- Constructor --- */

    /**
     * Constructor.
     *
     * @param za       the first point
     * @param zb       the second point
     */
    HTGeodesic(HTCoordE za, HTCoordE zb) {
        this.za    = za;
        this.zb    = zb;

        npoint  = new HTCoordE();
        zc = new HTCoordE();
        zo = new HTCoordE();
        a  = new HTCoordS();
        b  = new HTCoordS();
        c  = new HTCoordS();

        rebuild();
    }


  /* --- Refresh --- */

    /**
     * Refresh the screen coordinates of this node.
     *
     * @param sOrigin   the origin of the screen plane
     * @param sMax      the (xMax, yMax) point in the screen plane
     */
    void refreshScreenCoordinates(HTCoordS sOrigin, HTCoordS sMax) {
        a.projectionEtoS(za, sOrigin, sMax);
        b.projectionEtoS(zb, sOrigin, sMax);
        c.projectionEtoS(zc, sOrigin, sMax);
    }

  


  /* --- Rebuild --- */

    /**
     * Builds the geodesic.
     */
    void rebuild() {
        if ( (Math.abs(za.d()) < EPSILON) ||                       // za == origin
             (Math.abs(zb.d()) < EPSILON) ||                       // zb == origin
             (Math.abs((za.x / zb.x) - (za.y / zb.y)) < EPSILON) ) // za = lambda.zb
        {    
            type = LINE;
        } else {
            type = ARC;

            double da = 1 + za.x * za.x + za.y * za.y;
            double db = 1 + zb.x * zb.x + zb.y * zb.y;
            double dd = 2 * (za.x * zb.y - zb.x * za.y);
 
            zo.x = (zb.y * da - za.y * db) / dd;
            zo.y = (za.x * db - zb.x * da) / dd;

            double det = (zb.x - zo.x) * (za.y - zo.y) - (za.x - zo.x) * (zb.y - zo.y);
            double fa  = za.y * (za.y - zo.y) - za.x * (zo.x - za.x);
            double fb  = zb.y * (zb.y - zo.y) - zb.x * (zo.x - zb.x);

            zc.x = ((za.y - zo.y) * fb - (zb.y - zo.y) * fa) / det;
            zc.y = ((zo.x - za.x) * fb - (zo.x - zb.x) * fa) / det; 
        }
    }


    /* --- Draw --- */
    /**
     * Draws this geodesic.
     *
     * @param g    the graphic context
     */
    void draw(Graphics g, boolean fastMode, String cl1) {

		colorNode = new Color( Color.decode(cl1).getRGB() );  

		if ((fastMode) || (a.getDistance(b) < 33))
			type =  LINE;
		else
			type = ARC;
                            		
            switch(type) {
            	case LINE:		
                		g.drawLine(a.x, a.y, b.x, b.y);
                		break;
            	case ARC:		

                		//drawCurvedLine(g, a.x, a.y, c.x, c.y, b.x, b.y);
				color = g.getColor();

				drawCurvedLine(g, a.x, a.y, c.x, c.y, b.x, b.y);

				drawCurvedLine(g, a.x, a.y, c.x+1, c.y, b.x+1, b.y);
				drawCurvedLine(g, a.x, a.y, c.x-1, c.y, b.x-1, b.y);				

				drawCurvedLine(g, a.x, a.y, c.x, c.y+1, b.x, b.y+1);
				drawCurvedLine(g, a.x, a.y, c.x, c.y-1, b.x, b.y-1);

				drawCurvedLine(g, a.x, a.y, c.x+1, c.y, b.x+2, b.y);
				drawCurvedLine(g, a.x, a.y, c.x-1, c.y, b.x-2, b.y);				

				drawCurvedLine(g, a.x, a.y, c.x, c.y+1, b.x, b.y+2);
				drawCurvedLine(g, a.x, a.y, c.x, c.y-1, b.x, b.y-2);

	
                		break;
            	default:
                		break;
            }        
    } 

    /**
     * Draws this geodesic.
     *
     * @param g    the graphic context
     */

    void draw(Graphics g, boolean linhaDupla, boolean fastMode, String cl1) {

		colorNode = new Color( Color.decode(cl1).getRGB());  

		if ((fastMode) || (a.getDistance(b) < 33))

			type =  LINE;
		else
			type = ARC;		
                                		
            switch(type) {
            case LINE:
		    if (linhaDupla) {			
			g.drawLine(a.x+1, a.y, b.x+1, b.y);
			g.drawLine(a.x-1, a.y, b.x-1, b.y);					
			g.drawLine(a.x, a.y+1, b.x, b.y+1);
			g.drawLine(a.x, a.y-1, b.x, b.y-1);			
		    }		    
		    g.drawLine(a.x, a.y, b.x, b.y);		     
                break;
            case ARC:
		    if (linhaDupla) {  		
			drawCurvedLine1(g, a.x+1, a.y, c.x+1, c.y, b.x+1, b.y);
			drawCurvedLine1(g, a.x-1, a.y, c.x-1, c.y, b.x-1, b.y);	
			drawCurvedLine1(g, a.x+2, a.y, c.x+2, c.y, b.x+2, b.y);
			drawCurvedLine1(g, a.x-2, a.y, c.x-2, c.y, b.x-2, b.y);	
			drawCurvedLine1(g, a.x, a.y+1, c.x, c.y+1, b.x, b.y+1);
			drawCurvedLine1(g, a.x, a.y-1, c.x, c.y-1, b.x, b.y-1);	
			drawCurvedLine1(g, a.x, a.y+2, c.x, c.y+2, b.x, b.y+2);
			drawCurvedLine1(g, a.x, a.y-2, c.x, c.y-2, b.x, b.y-2);	
		    }	 		   		    
                drawCurvedLine1(g, a.x, a.y, c.x, c.y, b.x, b.y);	
                break;
            default:
                break;
            }
        
    } 

 
  void drawCurvedLine(Graphics g, double xi, double yi, double cx, double cy, double xf, double yf)
  {
	setValues(xi, yi, cx, cy, xf, yf);

	if (computeValues()) // gives new equation  
		myDoubleDrawPolygon(g);
	else 
		g.drawLine((int) xcoord[0] , (int) ycoord[0], (int) xcoord[2], (int) ycoord[2]);
  }

 void drawCurvedLine1(Graphics g, double xi, double yi, double cx, double cy, double xf, double yf)
  {
	setValues(xi, yi, cx, cy, xf, yf);

	if (computeValues()) // gives new equation  
		myDoubleDrawPolygon1(g);
	else 
		g.drawLine((int) xcoord[0] , (int) ycoord[0], (int) xcoord[2], (int) ycoord[2]);
  }

  void setValues (double xi, double yi, double cx, double cy, double xf, double yf) {
  	xcoord[0] = (double) xi;
    	ycoord[0] = (double) yi;
    	xcoord[1] = (double) cx;
    	ycoord[1] = (double) cy;
    	xcoord[2] = (double) xf;
    	ycoord[2] = (double) yf;
  }



 boolean computeValues2() { // computes values    

    double x1, y1, x2, y2;
    HTCoordE ps;
    double zax, zay, zbx, zby;

    zax = xValues[0];
    zay = yValues[0];
    zbx = xValues[2];
    zby = yValues[2];

    double da = 1 + zax * zax + zay * zay;
    double db = 1 + zbx * zbx + zby * zby;
    double dd = 2 * (zax * zby - zbx * zay);
 
    double zox = (zby * da - zay * db) / dd;
    double zoy = (zax * db - zbx * da) / dd;

    double det = (zbx - zox) * (zay - zoy) - (zax - zox) * (zby - zoy);
    double fa  = zay * (zay - zoy) - zax * (zox - zax);
    double fb  = zby * (zby - zoy) - zbx * (zox - zbx);

    // Preenchendo ponto faltantes

    xValues[0] = xcoord[0];
    yValues[0] = ycoord[0];

    xValues[5] = xcoord[1];
    yValues[5] = ycoord[1];

    xValues[numberSubintervals-1] = xcoord[2];
    yValues[numberSubintervals-1] = ycoord[2];


    ps = calcule(xValues[0], yValues[0], xValues[5], yValues[5], zox, zoy,  fb, fa, det);
    xValues[2]                  = ps.x;
    yValues[2]                  = ps.y;

    ps = calcule(xValues[5], yValues[5], xValues[10], yValues[10], zox, zoy,  fb, fa, det);
    xValues[8]                  = ps.x;
    yValues[8]                  = ps.y;


    ps = calcule(xValues[0], yValues[0], xValues[2], yValues[2], zox, zoy,  fb, fa, det);
    xValues[1]                  = ps.x;
    yValues[1]                  = ps.y;


    ps = calcule(xValues[8], yValues[8], xValues[10], yValues[10], zox, zoy,  fb, fa, det);
    xValues[9]                  = ps.x;
    yValues[9]                  = ps.y;


    x1                  	= (xValues[2] + xValues[5]) / 2;
    y1                  	= (yValues[2] + yValues[5]) / 2;

    x2                  	= (xValues[5] + xValues[8]) / 2;
    y2                  	= (yValues[5] + yValues[8]) / 2;

    ps = calcule(x1, y1, xValues[5], yValues[5], zox, zoy,  fb, fa, det);

    xValues[4]                  = ps.x;
    yValues[4]                  = ps.y;

    ps = calcule(x2, y2, xValues[8], yValues[8], zox, zoy,  fb, fa, det);

    xValues[6]                  = ps.x;
    yValues[6]                  = ps.y;

    ps = calcule(xValues[2], yValues[2], xValues[4], yValues[4], zox, zoy,  fb, fa, det);
    xValues[3]                  = ps.x;
    yValues[3]                  = ps.y;


    ps = calcule(xValues[6], yValues[6], xValues[8], yValues[8], zox, zoy,  fb, fa, det);
    xValues[7]                  = ps.x;
    yValues[7]                  = ps.y;
    
    xValues[numberSubintervals] = xcoord[2];
    yValues[numberSubintervals] = ycoord[2];

    return true;
  }

 

  boolean computeValues() { // computes values    

    double x1, y1, x2, y2;

    // Preenchendo ponto faltantes

    xValues[0] = xcoord[0];
    yValues[0] = ycoord[0];

    xValues[5] = xcoord[1];
    yValues[5] = ycoord[1];

    xValues[numberSubintervals-1] = xcoord[2];
    yValues[numberSubintervals-1] = ycoord[2];

    xValues[2]                  = (xValues[0] + xValues[5]) / 2;
    yValues[2]                  = (yValues[0] + yValues[5]) / 2;

    xValues[8]                  = (xValues[5] + xValues[10]) / 2;
    yValues[8]                  = (yValues[5] + yValues[10]) / 2;

    xValues[1]                  = (xValues[0] + xValues[2]) / 2;
    yValues[1]                  = (yValues[0] + yValues[2]) / 2;

    xValues[9]                  = (xValues[8] + xValues[10]) / 2;
    yValues[9]                  = (yValues[8] + yValues[10]) / 2;

    x1                  	= (xValues[2] + xValues[5]) / 2;
    y1                  	= (yValues[2] + yValues[5]) / 2;

    x2                  	= (xValues[5] + xValues[8]) / 2;
    y2                  	= (yValues[5] + yValues[8]) / 2;

    xValues[4]                  = (x1 + xValues[5]) / 2;
    yValues[4]                  = (y1 + yValues[5]) / 2;

    xValues[6]                  = (x2 + xValues[5]) / 2;
    yValues[6]                  = (y2 + yValues[5]) / 2;

    suavizaCentro1();

    suavizaCentro1();

    suavizaCentro1();

    suavizaTotal();

    suavizaTotal();

    suavizaTotal();

    //suavizaCentro();
    
    xValues[numberSubintervals] = xcoord[2];
    yValues[numberSubintervals] = ycoord[2];

    return true;
  }

  void suavizaCentro1() {
    xValues[3]                  = (xValues[2] + xValues[4]) / 2;
    yValues[3]                  = (yValues[2] + yValues[4]) / 2;

    xValues[7]                  = (xValues[6] + xValues[8]) / 2;
    yValues[7]                  = (yValues[6] + yValues[8]) / 2;

    xValues[5]                  = (xValues[4] + xValues[6]) / 2;
    yValues[5]                  = (yValues[4] + yValues[6]) / 2;

    xValues[4]                  = (xValues[3] + xValues[5]) / 2;
    yValues[4]                  = (yValues[3] + yValues[5]) / 2;

    xValues[6]                  = (xValues[5] + xValues[7]) / 2;
    yValues[6]                  = (yValues[5] + yValues[7]) / 2;

    xValues[2]                  = (xValues[1] + xValues[3]) / 2; //
    yValues[2]                  = (yValues[1] + yValues[3]) / 2; //

    xValues[7]                  = (xValues[6] + xValues[8]) / 2; //
    yValues[7]                  = (yValues[6] + yValues[8]) / 2; //

    xValues[5]                  = (xValues[4] + xValues[6]) / 2;
    yValues[5]                  = (yValues[4] + yValues[6]) / 2;

    xValues[4]                  = (xValues[3] + xValues[5]) / 2;
    yValues[4]                  = (yValues[3] + yValues[5]) / 2;

    xValues[6]                  = (xValues[5] + xValues[7]) / 2;
    yValues[6]                  = (yValues[5] + yValues[7]) / 2;

    //xValues[5]                  = (xValues[4] + xValues[6]) / 2;
    //yValues[5]                  = (yValues[4] + yValues[6]) / 2;
  }


  void suavizaTotal()
  {
    xValues[1]                  = (xValues[0] + xValues[2]) / 2;
    yValues[1]                  = (yValues[0] + yValues[2]) / 2;
    xValues[2]                  = (xValues[1] + xValues[3]) / 2;
    yValues[2]                  = (yValues[1] + yValues[3]) / 2;
    suavizaCentro();
    xValues[8]                  = (xValues[7] + xValues[9]) / 2;
    yValues[8]                  = (yValues[7] + yValues[9]) / 2;
    xValues[9]                  = (xValues[8] + xValues[10]) / 2;
    yValues[9]                  = (yValues[8] + yValues[10]) / 2;
  }

 void suavizaCentro()
  { 
    xValues[3]                  = (xValues[2] + xValues[4]) / 2;
    yValues[3]                  = (yValues[2] + yValues[4]) / 2;
    xValues[4]                  = (xValues[3] + xValues[5]) / 2;
    yValues[4]                  = (yValues[3] + yValues[5]) / 2;
    xValues[5]                  = (xValues[4] + xValues[6]) / 2;
    yValues[5]                  = (yValues[4] + yValues[6]) / 2;
    xValues[6]                  = (xValues[5] + xValues[7]) / 2;
    yValues[6]                  = (yValues[5] + yValues[7]) / 2;
    xValues[7]                  = (xValues[6] + xValues[8]) / 2;
    yValues[7]                  = (yValues[6] + yValues[8]) / 2;
  }
  
  void myDoubleDrawPolygon(Graphics g) {  

      Color c1 = color;
      Color c2 = colorNode;
       
       
    for (int i=0;  i < (numberSubintervals-1) ; i++) {     	

         float ratio = (float) i / (float) (numberSubintervals-1);
         int red   = (int)(c2.getRed() * ratio + c1.getRed() * (1 - ratio));
         int green = (int)(c2.getGreen() * ratio +  c1.getGreen() * (1 - ratio));
         int blue  = (int)(c2.getBlue() * ratio + c1.getBlue() * (1 - ratio));
         g.setColor(new Color(red, green, blue));        
	
      	g.drawLine((int) xValues[i], (int) yValues[i], (int) xValues[i+1], (int) yValues[i+1]);
    }    
  }

  void myDoubleDrawPolygon1(Graphics g) {  

       
    for (int i=0;  i < (numberSubintervals-1) ; i++) {     		
      	g.drawLine((int) xValues[i], (int) yValues[i], (int) xValues[i+1], (int) yValues[i+1]);
    }    
  }


  HTCoordE calcule(double zax, double zay, double zbx, double zby, double zox, double zoy,  double fb, double fa, double det) {

            npoint.x = ((zay - zoy) * fb - (zby - zoy) * fa) / det;
            npoint.y = ((zox - zax) * fb - (zox - zbx) * fa) / det; 

	    return npoint;
  }

  boolean computeValues1() {


	double gapX = (double) (xcoord[1] - xcoord[0])/ (double) numberSubintervals;
	double gapY = (double) (ycoord[1] - ycoord[0])/ (double) numberSubintervals;

    	xValues[0] = xcoord[0];
    	yValues[0] = ycoord[0];
   	xValues[numberSubintervals-1] = xcoord[2];
    	yValues[numberSubintervals-1] = ycoord[2];

	boolean pars = calcParameters();

	if (!pars)
		return false;	

	for(int i=1;i < numberSubintervals-1; i++) {
		if (gapX != 0) {
			xValues[i] = xcoord[0] + gapX*i;
			yValues[i] = PA * xValues[i] *xValues[i]  + PB * xValues[i]  + PC;

			
		}
		else {
			if (gapY != 0) {
				xValues[i] = xcoord[0];
				yValues[i] = yValues[i] + gapY;
			}
			else {
				yValues[i] = ycoord[0];
				xValues[i] = xValues[i] + gapX;
			}
		}
	}
	return true;

  }
boolean calcParameters() {

	double X1, X2, X3, Y1, Y2, Y3;

	X1 = xcoord[0];
	X2 = xcoord[1];
	X3 = xcoord[2];
	Y1 = ycoord[0];
	Y2 = ycoord[1];
	Y3 = ycoord[2];

	/*

	Solving for A gives:

    		A = [(Y2-Y1)(X1-X3) + (Y3-Y1)(X2-X1)]/[(X1-X3)(X2^2-X1^2) + (X2-X1)(X3^2-X1^2)]

		Plugging into the previous equation gives:

    		B = [(Y2 - Y1) - A(X2^2 - X1^2)] / (X2-X1)

		Plugging into the original equation gives:

    		C = Y1 - AX1^2 - BX1

	*/

	try {

		PA = ((Y2-Y1)*(X1-X3) + (Y3-Y1)*(X2-X1))/((X1-X3)*(X2*X2-X1*X1) + (X2-X1)*(X3*X3-X1*X1));

		PB = ((Y2 - Y1) - PA*(X2*X2 - X1*X1)) / (X2-X1);

		PC =  Y1 - PA*X1*X1 - PB*X1;

		return true; 
	} catch(Exception E) {
		System.out.println(E.toString());
		return false;
	}

  }


}

